import pandas as pd
import numpy as np


class ESTIMATE_signature:
    def __init__(self):
        self.sig = pd.DataFrame({"stromal": ["DCN",	"PAPPA",	"SFRP4",	"THBS2",	"LY86",	"CXCL14",	"FOXF1",	"COL10A1",	"ACTG2",	"APBB1IP",	"SH2D1A",	"SULF1",	"MSR1",	"C3AR1",	"FAP",	"PTGIS",	"ITGBL1",	"BGN",	"CXCL12",	"ECM2",	"FCGR2A",	"MS4A4A",	"WISP1",	"COL1A2",	"MS4A6A",	"EDNRA",	"VCAM1",	"GPR124",	"SCUBE2",	"AIF1",	"HEPH",	"LUM",	"PTGER3",	"RUNX1T1",	"CDH5",	"PIK3R5",	"RAMP3",	"LDB2",	"COX7A1",	"EDIL3",	"DDR2",	"FCGR2B",	"LPPR4",	"COL15A1",	"AOC3",	"ITIH3",	"FMO1",	"PRKG1",	"PLXDC1",	"VSIG4",	"COL6A3",	"SGCD",	"COL3A1",	"F13A1",	"OLFML1",	"IGSF6",	"COMP",	"HGF",	"GIMAP5",	"ABCA6",	"ITGAM",	"MAF",	"ITM2A",	"CLEC7A",	"ASPN",	"LRRC15",	"ERG",	"CD86",	"TRAT1",	"COL8A2",	"TCF21",	"CD93",	"CD163",	"GREM1",	"LMOD1",	"TLR2",	"ZEB2",	"C1QB",	"KCNJ8",	"KDR",	"CD33",	"RASGRP3",	"TNFSF4",	"CCR1",	"CSF1R",	"BTK",	"MFAP5",	"MXRA5",	"ISLR",	"ARHGAP28",	"ZFPM2",	"TLR7",	"ADAM12",	"OLFML2B",	"ENPP2",	"CILP",	"SIGLEC1",	"SPON2",	"PLXNC1",	"ADAMTS5",	"SAMSN1",	"CH25H",	"COL14A1",	"EMCN",	"RGS4",	"PCDH12",	"RARRES2",	"CD248",	"PDGFRB",	"C1QA",	"COL5A3",	"IGF1",	"SP140",	"TFEC",	"TNN",	"ATP8B4",	"ZNF423",	"FRZB",	"SERPING1",	"ENPEP",	"CD14",	"DIO2",	"FPR1",	"IL18R1",	"HDC",	"TXNDC3",	"PDE2A",	"RSAD2",	"ITIH5",	"FASLG",	"MMP3",	"NOX4",	"WNT2",	"LRRC32",	"CXCL9",	"ODZ4",	"FBLN2",	"EGFL6",	"IL1B",	"SPON1",	"CD200"],
    "immune" :["LCP2",	"LSP1",	"FYB",	"PLEK",	"HCK",	"IL10RA",	"LILRB1",	"NCKAP1L",	"LAIR1",	"NCF2",	"CYBB",	"PTPRC",	"IL7R",	"LAPTM5",	"CD53",	"EVI2B",	"SLA",	"ITGB2",	"GIMAP4",	"MYO1F",	"HCLS1",	"MNDA",	"IL2RG",	"CD48",	"AOAH",	"CCL5",	"LTB",	"GMFG",	"GIMAP6",	"GZMK",	"LST1",	"GPR65",	"LILRB2",	"WIPF1",	"CD37",	"BIN2",	"FCER1G",	"IKZF1",	"TYROBP",	"FGL2",	"FLI1",	"IRF8",	"ARHGAP15",	"SH2B3",	"TNFRSF1B",	"DOCK2",	"CD2",	"ARHGEF6",	"CORO1A",	"LY96",	"LYZ",	"ITGAL",	"TNFAIP3",	"RNASE6",	"TGFB1",	"PSTPIP1",	"CST7",	"RGS1",	"FGR",	"SELL",	"MICAL1",	"TRAF3IP3",	"ITGA4",	"MAFB",	"ARHGDIB",	"IL4R",	"RHOH",	"HLA-DPA1",	"NKG7",	"NCF4",	"LPXN",	"ITK",	"SELPLG",	"HLA-DPB1",	"CD3D",	"CD300A",	"IL2RB",	"ADCY7",	"PTGER4",	"SRGN",	"CD247",	"CCR7",	"MSN",	"ALOX5AP",	"PTGER2",	"RAC2",	"GBP2",	"VAV1",	"CLEC2B",	"P2RY14",	"NFKBIA",	"S100A9",	"IFI30",	"MFSD1",	"RASSF2",	"TPP1",	"RHOG",	"CLEC4A",	"GZMB",	"PVRIG",	"S100A8",	"CASP1",	"BCL2A1",	"HLA-E",	"KLRB1",	"GNLY",	"RAB27A",	"IL18RAP",	"TPST2",	"EMP3",	"GMIP",	"LCK",	"IL32",	"PTPRCAP",	"LGALS9",	"CCDC69",	"SAMHD1",	"TAP1",	"GBP1",	"CTSS",	"GZMH",	"ADAM8",	"GLRX",	"PRF1",	"CD69",	"HLA-B",	"HLA-DMA",	"CD74",	"KLRK1",	"PTPRE",	"HLA-DRA",	"VNN2",	"TCIRG1",	"RABGAP1L",	"CSTA",	"ZAP70",	"HLA-F",	"HLA-G",	"CD52",	"CD302",	"CD27"]})
    def get_signature_data(self):
        return self.sig


def ESTIMATEscore(df):
    """
    Compute the ESTIMATE immune and stromal score for the input gene expression matrix
    df: the input gene expression matrix dataframe where gene symbol as the first column or index
    index: whether gene symbol column is in the index. True if yes, and no other gene symbol column in the dataframe; otherwise no
    output: a pandas dataframe with three columns, 
            index: sample ID, 
            'EST_stromal': ESTIMATE stromal score; 
            'EST_immune': ESTIMATE immune score; 
            'estimate': ESTIMATE score
    """

    df_ranked = df.apply(pd.to_numeric).rank(axis = 0, method = "average")
    df_ranked.index = df.index
    df_normalized = 10000*df_ranked/df_ranked.shape[0]
    gene_set_names = ['stromal', 'immune']
    score = np.empty((df_ranked.shape[1], 2))

    common_genes = {}
    signature_data = ESTIMATE_signature()
    sig = signature_data.get_signature_data()
    for i in range(0,2):
    
        gene_set = sig.iloc[:,i]
        common = list(set(gene_set) & set(df.index))
        common_genes[gene_set_names[i]] = common

        if len(common) == 0: 
            print("There is no common genes with the input and the ", gene_set_names[i], " signature")
            continue

        # ssGSEA
        ES_vector = [0]*df_normalized.shape[1]
        for sample in range(0, df_normalized.shape[1]):
            ordered_genes = df_normalized.iloc[:,sample].sort_values(ascending = False)
            hit_ind = ordered_genes.index.isin(gene_set)
            no_hit_ind = 1- hit_ind
            ordered_genes1 = ordered_genes.pow(1./4)
            hit_exp = ordered_genes1[hit_ind]
            no_hit_penalty = np.cumsum(no_hit_ind/sum(no_hit_ind))
            hit_reward = np.cumsum((hit_ind*ordered_genes1)/sum(hit_exp))
            ES_vector[sample] = sum(hit_reward - no_hit_penalty)
        score[:, i] = ES_vector

    score = pd.DataFrame(score)
    score.index = df.columns
    score.columns = ["EST_stromal", "EST_immune"]
    score['estimate'] = score['EST_stromal'] + score['EST_immune']

    return score



def tumor_purity(platform, df):
    """
    When data comes from platform Affymetrix, provides an estimate of tumor purity
    output: numpy array
    """
    if platform != "Affymetrix":
        raise ValueError("Tumor purity is not available for this platform")
    else:
        score = ESTIMATEscore(df)
        purity = np.cos(0.6049872018 + 0.0001467884*score['estimate'])
        return purity
    

def common_genes(df):
    """
    Output the common genes of the input df and the ESTIMATE signatures
    """
    genes = df.iloc[:,0]
    signature_data = ESTIMATE_signature()
    sig = signature_data.get_signature_data()
    common_immune = list(set(sig['immune']) & set(genes))
    common_stromal = list(set(sig['stromal']) & set(genes))
    common = {'EST_immune': common_immune, 'EST_stromal': common_stromal}
    return common


import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve, auc
from lifelines import KaplanMeierFitter
from lifelines.plotting import add_at_risk_counts
from lifelines import CoxPHFitter
from lifelines.statistics import logrank_test
from lifelines.utils import concordance_index
import pandas as pd
import numpy as np

def assign_type(score, upper_p, lower_p):
    if upper_p == lower_p:
        q = score.quantile(upper_p)
        def get_type(value):
            if value >= q:
                return 'H'
            else:
                return 'L'
    
    else:
        upper_q = score.quantile(upper_p)
        lower_q = score.quantile(lower_p)
        def get_type(value):
            if value >= upper_q:
                return 'H'
            elif value <= lower_q:
                return 'M'
            else:
                return 'L'
    
    return score.apply(get_type)



def optimal_ICI(df, response, score_name, name = None, roc = True):
    """
    Find the optimal score with the highest AUC for ICI response
    df: pandas dataframe with row index as sample ID, columns containing scores, ICI response ([R, NR] or [0, 1])
    response: column name of ICI response
    score_name: list of names of the score columns in df, or a string of name for only one score
    name: custom title name for figure
    output: a figure showing all ROC curves by scores, and return a dictionary with all scores and the corresponding AUC
    """

    # remove samples with no response status
    df1 = df[~df[response].isna()]
    if isinstance(score_name, str):
        score_name = [score_name]
    
    # transform the response column
    all_strings = df1[response].map(lambda x: isinstance(x, str)).all()
    if all_strings:
        df1[response] = df1[response].str.upper()
        if df1[response].isin(["R","NR"]).all():
            df1["binary_resp"] = df1[response].apply(lambda x: 1 if x == "R" else 0)
        else:
            raise ValueError("Unsupported response type")
    elif df1[response].isin([0,1]).all():
        df1["binary_resp"] = df1[response]
    else:
        raise ValueError("Unsupported response type")

    fig = plt.figure(figsize=(8, 6), dpi = 100)  # Create a figure for the plot
    plt.margins(x=0.05, y=0.05)
    legend_list = score_name
    auc_dict = {}

    for i in range(len(score_name)):
        # Remove rows with missing values in the current score column
        df2 = df1[~df1[score_name[i]].isna()]
        df2[score_name[i]] = df2[score_name[i]].astype(float)
        # Compute FPR, TPR, and thresholds for the ROC curve
        fpr, tpr, thresholds = roc_curve(
            df2['binary_resp'],
            df2[score_name[i]])
        
        # Compute the Area Under the Curve (AUC)
        roc_auc = auc(fpr, tpr)
        auc_dict[score_name[i]] = roc_auc
        
        # Plot the ROC curve
        if roc:
            plt.plot(fpr, tpr, lw=4, label=f'{legend_list[i]} ({roc_auc:.2f})')
            plt.title("ROC Curves for ICI Prediction", fontweight = "bold", fontsize = 16)
            plt.xlabel("False Positive Rates", fontweight = "bold", fontsize = 14)
            plt.ylabel("True Positive Rates", fontweight = "bold", fontsize = 14)
            plt.xticks(fontweight="bold", fontsize = 14)
            plt.yticks(fontweight="bold", fontsize = 14)

    # Plot the diagonal line (random classifier)
    if roc:
        plt.plot([0, 1], [0, 1], color='grey', lw=1, linestyle='--')

        # Set plot limits and labels
        plt.xlim([0.0, 1.03])
        plt.ylim([0.0, 1.1])
        plt.xticks(fontweight="bold", fontsize = 14)
        plt.yticks(fontweight="bold", fontsize = 14)

        if name is None:
            plt.title(f"ROC Curves", fontweight = "bold", fontsize = 16)
        else:
            plt.title(f"ROC Curves for {name}", fontweight = "bold", fontsize = 16) 

        plt.legend(loc="upper left", bbox_to_anchor=(0, 1), title = "AUC",
                   title_fontproperties={'weight': 'bold', 'size': 12}, 
                   prop = {'size':10, 'weight' : "bold"},
                   frameon = False)

    best_auc = max(auc_dict.values())
    optimal_score = [k for k, v in auc_dict.items() if v == best_auc]

    print(f"The optimal score for ICI response prediction: {optimal_score} with AUC = {round(best_auc, 4)}")

    return fig, auc_dict


def optimal_survival(df, delta, time, score_name, clinical_factors = None, upper_p = 0.75, lower_p = 0.25, name = None, km_curves = True, palette = ["#547AC0", "#898988", "#F6C957"]):
    """
    Compare score performance to survival prognosis using Cox-PH model. Each score is divided into three groups, and users can choose to add other clinical factors into the model to adjust for confounding variables.
    delta: column name for survival status. 1 as event (death) and 0 otherwise
    time: column name for survival time, should be float type
    score_name: list of names of the score columns in df, or a string of name for only one score
    clinical factors: list of column names of the other clinical factors added to the model
    upper_p, lower_p: the probabilities at which the upper and lower quantiles are calculated, by default 0.75 and 0.25, respectively
    name: custom title name for the figure
    km_curves: whether to display the Kaplan-Meier survival curve of the optimal score divided into three groups by upper_p and lower_p
    palette: palette for the K-M survival curves
    output: return a dictionary with all scores and their corresponding c-index
    """

    # only considers sample with known survival status and time
    df1 = df[~df[delta].isna()]
    df1 = df1[~df1[time].isna()]
    c_dict = {}

    # test if input values are valid
    if not df1[delta].isin([1,0]).all():
        raise ValueError("delta must be a column containing only 0 or 1")

    if isinstance(score_name, str):
        score_name = [score_name]
    
    if clinical_factors is not None:
        if isinstance(clinical_factors, str):
            clinical_factors = [clinical_factors]
        elif not isinstance(clinical_factors, list):
            raise TypeError("Invalid clinical factor input")
        
    # conduct Cox-PH model
    for score in score_name:
        if clinical_factors is None:
            df1[score] = df1[score].astype(float)
            c_ind = concordance_index(df1[time], df1[score], df1[delta])
        else:
            cph = CoxPHFitter()
            data = df[[delta, time, score] + clinical_factors]
            cph.fit(data, duration_col = time, event_col = delta)
            predicted_risk = cph.predict_partial_hazard(data)
            c_ind = concordance_index(data[time], predicted_risk, data[delta])
        
        c_dict[score] = c_ind
        
    # select the best method and highest c-index
    best_c = max(c_dict.values())
    optimal_score = [k for k, v in c_dict.items() if v == best_c]
    print(f"The optimal score for survival prognosis: {optimal_score} with c-index = {round(best_c, 4)}")

    # draw K-M survival curves
    if km_curves:
        fig, axs = plt.subplots(1, len(optimal_score), figsize=(8, 6*len(optimal_score)), 
                                dpi = 100) #constrained_layout=True, 

        if len(optimal_score) == 1:
            axs = [axs]
        for i in range(len(optimal_score)):
            # segment samples by score into three groups: high(H), and low(L)
            score_type = assign_type(df1[optimal_score[i]], upper_p, lower_p)
            df1['score_type'] = score_type
            kmf_H = KaplanMeierFitter()
            kmf_L = KaplanMeierFitter()
            kmf_H.fit(durations=df1[df1['score_type'] == 'H'][time], event_observed=df1[df1['score_type'] == 'H'][delta], label='H')
            kmf_H.plot_survival_function(show_censors = True, ci_show = False, linewidth=4, color = palette[0], ax = axs[i])
            # add the M group if there are two quantile probabilities
            if upper_p != lower_p:
                kmf_M = KaplanMeierFitter()
                kmf_M.fit(durations=df1[df1['score_type'] == 'M'][time], event_observed=df1[df1['score_type'] == 'M'][delta], label='M')
                kmf_M.plot_survival_function(show_censors = True, ci_show = False, linewidth=4, color = palette[2], ax = axs[i])

            kmf_L.fit(durations=df1[df1['score_type'] == 'L'][time], event_observed=df1[df1['score_type'] == 'L'][delta], label='L')
            kmf_L.plot_survival_function(show_censors = True, ci_show = False, linewidth=4, color = palette[1], ax = axs[i])
            # add risk table for high and low groups
            #add_at_risk_counts(kmf_H, kmf_L, ax=axs[i])

            # perform log-rank test
            df1[time] = pd.to_numeric(df1[time], errors='coerce')
            df1[delta] = pd.to_numeric(df1[delta], errors='coerce')
            results_HL = logrank_test(df1[df1['score_type'] == 'H'][time], df1[df1['score_type'] == 'L'][time], 
                                      event_observed_A=df1[df1['score_type'] == 'H'][delta], 
                                      event_observed_B=df1[df1['score_type'] == 'L'][delta])

            axs[i].margins(x=0.05, y=0.05)
            axs[i].text(0.5, 0.85, f"H vs L: P  = {round(results_HL.p_value,4)}", 
                        transform=axs[i].transAxes, verticalalignment='top', fontweight = 'bold', fontsize = 14)
            axs[i].text(0.5, 0.75, f"c-index = {round(best_c,4)}", transform=axs[i].transAxes, 
                        verticalalignment='top', fontweight = 'bold', fontsize = 14)
            
            if name is None:
                axs[i].set_title(f"K-M Curves for {optimal_score[i]}", fontweight = 'bold', fontsize = 16)
            else:
                axs[i].set_title(f"K-M Curves for {optimal_score[i]} in {name}", fontweight = 'bold', fontsize = 16)
            axs[i].set_xlabel(time, fontweight = 'bold', fontsize = 14)
            axs[i].set_ylabel("Survival Rate", fontweight = 'bold', fontsize = 14)

            # bold all axis
            xticks = axs[i].get_xticks()
            yticks = axs[i].get_yticks()
            xticks_rounded = np.round(xticks, decimals=0)
            yticks_rounded = np.round(yticks, decimals = 1)
            axs[i].set_xticklabels([f'{tick:.0f}' for tick in xticks_rounded], weight = 'bold', size = 14)
            axs[i].set_yticklabels([f'{tick:.1f}' for tick in yticks_rounded] ,weight = 'bold', size = 14)
            axs[i].legend(
                        loc="upper right", title="Group",
                        title_fontproperties={'weight': 'bold', 'size': 12},  # Bold legend title
                        prop={'size': 10, 'weight': "bold"}
                    )


    return fig, c_dict


def get_performance(df_score, metric, score_name, surv_col = None, surv_p = None, ICI_col = None, df_clin = None, clinical_factors = None, show_fig = True, name = None):
    """
    Compare score performance of ICI therapy or/and survival prognosis
    df_score: pandas dataframe containing scores with sample ID as row index
    metric: choose from ["survival", "ICI"]. Enter the full list if the user wants to compare both.
    surv_col: column names of the survival comparison, should be a list of [status, time]
    surv_p: quantile probability for score segmentation in K-M survival curves
    ICI_col: column name of the response column. If comparing both metrics, col = [metric1 col, metric2 col] corresponding to the method.
    score_name: list of names of the score columns in df, or a string of name for only one score
    df_clin: if the columns of survival or response are not in df_score, provide them as a column with sample ID as row index in the pandas dataframe
    clinical_factors: additional column names of factors adding into 
    show_fig: whether to show the ROC curves/K-M survival curves, default to True
    output: return a list of two dictionaries with all scores and their corresponding c-indices/AUCs
    """

    if isinstance(metric, str):
        metric = [metric]
    elif not isinstance(metric, list):
        raise TypeError("Invalid format, must be a list of metrics or a string of a single metric")
    
    df = df_score.copy()
    if df_clin is not None:
        df = df.merge(df_clin, left_index = True, right_index = True, how = "inner")
    
    # empty outcome list
    outcomes = []

    for m in metric:
        if m == "survival":
            # test whether the survival columns are valid
            if not isinstance(surv_col, list):
                raise TypeError("The input survival columns must be in a list")
            if len(surv_col)!= 2:
                raise ValueError("The input survival columns must only have status and time")
            # assigning delta and time to survival columns
            if df[surv_col[0]].dropna().isin([1,0]).all():
                delta = surv_col[0]
                time = surv_col[1]
            else:
                delta = surv_col[1]
                time = surv_col[0]

            # distinguish the upper and lower quantile probabilities and compute survival outcomes
            if surv_p is None:
                surv_out = optimal_survival(df, delta, time, score_name, clinical_factors = None, name = name, km_curves = show_fig)
            else:
                if isinstance(surv_p, float):
                    upper_p = surv_p
                    lower_p = surv_p
                elif not isinstance(surv_p, list):
                    raise TypeError("Invalid input format of quantile probabilities")
                else:
                    upper_p = max(surv_p)
                    lower_p = min(surv_p)
                surv_out = optimal_survival(df, delta, time, score_name, clinical_factors = None, upper_p = upper_p, lower_p = lower_p, name = name, km_curves = show_fig)
            outcomes.append(surv_out)
        # compute ICI outcome
        elif m == "ICI":
            if not isinstance(ICI_col, str):
                raise TypeError("The input ICI_col must be a string")
            elif ICI_col is None:
                raise ValueError("Must input a valid column for ICI therapy response")

            ICI_out = optimal_ICI(df, ICI_col, score_name, name = name, roc = show_fig)
            outcomes.append(ICI_out)

        else:
            raise ValueError("Invalid metric, must be survival or ICI")
        
    return outcomes
        
    



import pandas as pd
import numpy as np
import warnings

class ISTME_signature:
    def __init__(self):
        self.immune = ['CD40LG','THEMIS','PYHIN1','TRAT1','FCRL1','SPIB','GPR174','SH2D1A','CCR4','ITK','FCRL3','C5orf20','FIGF','UBASH3A','MS4A1','CD300LG','SAMD3','ADH1B','GZMK','CHRDL1','ABI3BP','FCER2','TIFAB','P2RY12','CLEC10A','RSPO2','PCDH15','HLA-DOA','PLA2G2D','CLEC17A','CD3G','CCL19','PTPRC','C17orf87','GRIA1','CD8A','PRG4','P2RY13','SFTPC','HLA-DPB1','AADAC','EOMES','AOAH','CD1E','CCR2','CCL5','GFRA1','TFEC','CLDN18','FGL2','C4orf7','CD1B','GZMA','HLA-DPA1','SCARA5','PLEK','ZNF683','CD19','HLA-DRA','CD84','PIK3CG','NCKAP1L','CR1','WIF1','CLEC12A','KIAA0408','PIGR','CXCL13','CD74','TLR8','CHIT1','IL7R','HLA-DRB1','HLA-DQA1','COL6A6','LYZ','SFTPA2','CYBB','COL29A1','LCP1','IGJ','UBD','HLA-DQA2','CXCL9','PPYR1','IFNG','MRC1','FAM26F','B2M','CD1A','HLA-B','GBP5','ADAMDEC1','LAPTM5','C1QB','ITGB2','C1QA','SUCNR1','MARCO','HLA-C','VSIG4','F13A1','CXCL10','C1QC','HLA-A','CD163','FCGR3A','TMSL3']
        self.stromal = ['MDK','RPL8','S100P','FTH1','GPR87','UBC','CLDN3','FGL1','TMSB10','HSPB1','ACTB','ITPKA','KRT19','S100A2','EEF1A2','PABPC1','RPLP0','HSP90AB1','NMU','KRT8','KRT7','EPCAM','YWHAZ','MYH9','ACTG1','NPW','MMP13','TUBA1B','ERRFI1','DSP','KRT18','P4HB','ENO1','PKM2','RHOV','ALDOA','JUP','LDHA','HMGA1','HSPA1A','MIF','TUBB','HMGB3','CYP24A1','TPI1','FAM83A','RECQL4','FN1','VEGFA','GAPDH','TK1','TUBB3','UBE2C','MYBL2','COL3A1','COL1A1','COL1A2','COL11A1']
    def get_signature_data(self):
        geneset = {"immune": self.immune, "stromal": self.stromal}
        return geneset

def istmeScore(df):
    """
    Perform ssGSEA and compute the ISTME stromal and immune score
    df: normalized gene expression matrix, with gene symbol as the first column or index
    index: whether gene symbols are the row indices of df
    output: a pandas dataframe with two columns, 
        index: sample ID; 
        ISTME_immune: ISTME immune score; 
        ISTME_stromal: ISTME stromal score
    """
    signature = ISTME_signature()
    geneset = signature.get_signature_data()

    df = df.apply(pd.to_numeric)
    df_ranked = df.rank(axis = 0, method = "average")
    df_ranked = df_ranked.apply(abs)
    score = np.empty((df_ranked.shape[1], 2))
    ES_vector = [0]*df.shape[1]
    i = 0

    for sig in geneset.keys():

        for sample in range(0, df.shape[1]):
            ordered_genes = df_ranked.iloc[:,sample].sort_values(ascending = False)
            hit_ind = ordered_genes.index.isin(geneset[sig])
            no_hit_ind = 1- hit_ind
            ordered_genes1 = ordered_genes.pow(1./4)
            hit_exp = ordered_genes1[hit_ind]
            no_hit_penalty = np.cumsum(no_hit_ind/sum(no_hit_ind))
            hit_reward = np.cumsum((hit_ind*ordered_genes1)/sum(hit_exp))
            ES_vector[sample] = sum(hit_reward - no_hit_penalty)

        ES_normalized = [i/(max(ES_vector)-min(ES_vector)) for i in ES_vector]
        score[:,i] = ES_normalized
        i += 1

    score = pd.DataFrame(score)
    score.index = df_ranked.columns
    score.columns = ['IS_stromal', 'IS_immune']

    return score


def get_subtypes(score, immune_q = 0.5, stromal_q = 0.5):
    """
    Get four TME subtypes
    """
    warnings.filterwarnings('ignore')
    score["group"] = [0]*score.shape[0]
    immune_quantile = score['immune'].quantile(q = immune_q)
    stromal_quantile = score['stromal'].quantile(q = stromal_q)

    score.loc[(score['immune'] <= immune_quantile) & (score['stromal'] <= stromal_quantile).to_list(),'group'] = "immune_L, stromal_L"
    score.loc[(score['immune'] > immune_quantile) & (score['stromal'] <= stromal_quantile).to_list(),'group'] = "immune_H, stromal_L"
    score.loc[(score['immune'] <= immune_quantile) & (score['stromal'] > stromal_quantile).to_list(),'group'] = "immune_L, stromal_H"
    score.loc[(score['immune'] > immune_quantile) & (score['stromal'] > stromal_quantile).to_list(),'group'] = "immune_H, stromal_H"

    return(score)



import gseapy as gp
import pandas as pd
import requests
import numpy as np
import json

class nb_pathway():
    def __init__(self):
        url = "https://raw.githubusercontent.com/qiluzhou/TMEImmune/refs/heads/main/data/"
        self.coef = pd.read_csv(url + "coef_best_netbio.csv")
        response = requests.get(url + "gene_sets.json")
        if response.status_code == 200:
            self.gene_set = json.loads(response.text)
        else:
            print(f"Failed to retrieve the reactomr geneset: {response.status_code}")
    
    def get_nb_coef(self):
        return self.coef
    
    def reactome_geneset(self):
        return self.gene_set



def get_netbio(df):
    """
    Compute the NetBio score for the input gene expression matrix
    df: the input gene expression matrix dataframe where gene symbol as the first column or index
    index: whether gene symbol column is in the index. True if yes, and no other gene symbol column in the dataframe; otherwise no
    output: a pandas dataframe with one column, 
            index: sample ID, 
            'NetBio_score': NetBio score
    """

    # coefficients of netbio pathways
    nb_coef = nb_pathway().get_nb_coef()
    pathway = nb_coef['Variable'][1:]

    # get netbio reactome pathways and corresponding genesets
    gene_set_dict = nb_pathway().reactome_geneset()

    # perform ssgsea
    ssgsea_results = gp.ssgsea(
    data=df,
    gene_sets=gene_set_dict,  # Path to the gene set file
    min_size=0,
    outdir=None,  # Avoid file output
    verbose=True
    )
    nes_pivot = ssgsea_results.res2d.pivot(index='Term', columns='Name', values='NES')

    # find the common pathways of ssgsea results and the netbio pathways
    common_path = list(set(pathway) & set(nes_pivot.index))
    nes_common = nes_pivot.loc[list(common_path),:]
    nes_common = nes_common.astype(float)
    # replace NA with 0
    nes_common = nes_common.fillna(0)
    ptys = nes_common.index

    # extract coefficients of the common pathways
    nb_coef1 = nb_coef[nb_coef['Variable'].isin(common_path)]

    # arrange the coefficients as the same order as the ssgsea result dataframe
    nb_coef1['Variable'] = pd.Categorical(nb_coef1['Variable'], categories=list(ptys), ordered=True)
    nb_coef_sorted = nb_coef1.sort_values('Variable')
    coefficients = nb_coef_sorted['Coefficient']

    # calculate the netbio score
    intercept = nb_coef['Coefficient'][0]
    nb_score = np.dot(nes_common.T, coefficients) + intercept
    netbio = pd.DataFrame({"NetBio": nb_score.astype(float)})
    netbio.index = nes_pivot.columns

    return netbio


import pandas as pd
import numpy as np
import os
from cmapPy.pandasGEXpress.parse_gct import parse
from rnanorm import TMM
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from inmoose.pycombat import pycombat_seq
import warnings



def read_data(path):
    """
    Format: supports txt, csv, gct input
    Input: gene expression matrix with gene symbol as the first column, samples in columns and gene symbols in rows. 
           Duplicated or missing genes are not allowed.
    Output: a pandas dataframe where gene symbols are the row index and samples are columns
    """
    # gct, txt, csv, first column as genes
    _, file_extension = os.path.splitext(path)
    if file_extension == '.txt':
        df = pd.read_table(path, sep = "\t", index_col=0) 

    elif file_extension == '.csv':
        df = pd.read_csv(path, index_col=0)

    elif file_extension == '.gct':
        df = parse(path)
        df = df.data_df

    else:
        raise TypeError(file_extension + " file is not supported")

    genes = df.index
    if genes.duplicated().any():
        dup_genes = genes[genes.duplicated()]
        warn = str("Duplicate genes: " + dup_genes.unique())
        warnings.warn(warn, category=UserWarning)
    
    if genes.isnull().any():
        warnings.warn("Exist NA's in the input genes", category=UserWarning)

    return (df)


def normalization(df = None, path = None, method = None, minmax = False, zscore = False, batch = None, batch_col = None): 
    """
    Perform normalization and batch effect correction to the input. The input must be a pandas dataframe or a valid path.
    Method: choose method from TMM, CPM, median ratio for read count data. Default to None for normalized gene expression.
    minmax: perform min-max normalization, set to False by default
    zscore: perform z-score normalization, set to False by default
    batch: a pandas dataframe having batch information for the gene expression data to perform batch effect correction, with row index as sample ID
    batch_col: column name of the batch information, must be a string
    """

    # read data from path if path exists
    if path is not None:
        df = read_data(path)
    else:
        if df is None:
            raise ValueError("Invalid input pandas dataframe and path")
        elif not isinstance(df, pd.DataFrame):
            raise TypeError("The input should be a pandas dataframe")

    # test whether the first column is genes
    first_col = df.iloc[:,0]
    has_letters = first_col.apply(lambda x: any(char.isalpha() for char in str(x)))
    if has_letters.all():
        df1 = df.copy()
        df1.index = first_col
        df1 = df1.iloc[:,1:]
    else:
        ind_letters = df.index.to_series().apply(lambda x: any(char.isalpha() for char in str(x)))
        if not ind_letters.any():
            raise ValueError("Index contains invalid gene name. Gene symbols must be in the first column or row index")
        else:
            df1 = df.copy()

    # fill NA's with 0 count
    df1 = df1.fillna(0)

    # read count normalization
    if method is not None:

        if method == 'TMM':
            # calculate TMM
            norm = TMM().fit(df1.T)
            gene_norm = norm.transform(df1.T)
            gene_norm = pd.DataFrame(gene_norm).T
            gene_norm.columns = df1.columns
            gene_norm.index = df1.index
            print("---------- data has been TMM normalized ----------")
        elif method == 'CPM':
            column_sums = df1.sum(axis=0)
            # calculate CPM
            gene_norm = (df1 / column_sums) * 1e6
            gene_norm.index = df1.index
            print("---------- data has been CPM normalized ----------")
        elif method == 'median ratio':
            # compute geometric means of each gene
            geo_means = np.exp(np.log(df1.replace(0, np.nan)).mean(axis=1))
            ratios = df1.div(geo_means, axis=0)
            # calculate the median of the ratios for each sample
            medratio = ratios.median(axis=0)
            medratio1 = medratio.replace(0, 1e-06)
            gene_norm = df1.div(medratio1, axis=1)
            gene_norm.index = df1.index
            print("---------- data has been median ratio normalized ----------")
        else:
            raise TypeError("Normalization method not supported")
        
    df_norm = gene_norm if 'gene_norm' in locals() else df1
    # remove incorrectly formatted rows
    df_norm = df_norm[~pd.to_datetime(df_norm.index, errors='coerce').notna()]

    # z-score normalization
    if zscore:
        scalar = StandardScaler()
        df_norm = pd.DataFrame(scalar.fit_transform(df_norm.T).T,  # Transpose to scale across rows (genes)
                                  index=df_norm.index, columns=df_norm.columns)
        print("---------- data has been z-score normalized ----------")

    # min-max normalization
    if minmax:
        min_max_scaler = MinMaxScaler(feature_range=(0, 1))  # Adjust range as needed
        df_norm = pd.DataFrame(min_max_scaler.fit_transform(df_norm.T).T,  # Transpose to scale across rows (genes)
                                        index=df_norm.index, columns=df_norm.columns)
        print("---------- data has been min-max normalized ----------")
    

    if batch is not None:
        if not isinstance(batch_col, str):
            raise TypeError("batch_col must be a string")
        # replace negative expression values to the absolute values
        df_norm = df_norm.abs()
        warnings.warn("Negative gene expression exists in the dataframe", category=UserWarning)

        # reorder batch by the order of the gene expression matrix columns
        common_ID = set(df_norm.columns) & set(batch.index)
        df_norm = df_norm[list(common_ID)]
        common_batch = batch.loc[list(common_ID),:]
        batch_ordered = common_batch.loc[df_norm.columns].reset_index()

        # remove batch with only one sample and raise warning
        unique_batch = batch_ordered[batch_col].value_counts()[batch_ordered[batch_col].value_counts() == 1].index
        
        if not unique_batch.empty:
            warnings.warn(f"The following batches with only one sample will be removed: {list(unique_batch)}")
            sample_id = batch[batch[batch_col].isin(unique_batch)].index
            single_batch_colname = list(set(sample_id) & set(df_norm.columns))
            batch_ordered = batch_ordered[~batch_ordered[batch_col].isin(unique_batch)]
            df_norm = df_norm.drop(columns = single_batch_colname)
        df_norm = pycombat_seq(df_norm, batch_ordered[batch_col])
        print("---------- batch effect removed ----------")
            
    return df_norm


from TMEImmune import estimateScore, netbio, SIAscore
from TMEImmune import ISTME as ISM
import pandas as pd
import warnings

def get_score(df, method):
    """
    Compute TME scores for the input gene expression data.
    df: a pandas dataframe having gene symbol as the first column or row index
    method: TME scoring methods choosing from ESTIMATE, ISTME, NetBio, SIA. The method statement can be one method, or a list of multiple methods.
    Output: a pandas dataframe with sample ID as index, and columns are the computed scores
    """

    # test if input is a pandas dataframe
    if not isinstance(df, pd.DataFrame):
        raise TypeError("Input must be a pandas dataframe")


    # set gene symbol as row index if not the index column
    first_col = df.iloc[:,0]
    has_letters = first_col.apply(lambda x: any(char.isalpha() for char in str(x)))
    if has_letters.all():
        df1 = df.copy()
        df.index = first_col
        df = df.iloc[:,1:]
    else:
        ind_letters = df.index.to_series().apply(lambda x: any(char.isalpha() for char in str(x)))
        if not ind_letters.any():
            raise ValueError("Index contains invalid gene name. Gene symbols must be in the first column or row index")
        else:
            df1 = df.copy()

    # convert non-float values to float
    all_floats = df1.map(lambda x: isinstance(x, float)).all().all()
    if not all_floats:
        try:
            df1 = df1.astype(float)
        except ValueError:
            raise ValueError("Input contains non-float convertible values.")    

    # remove name missing genes
    if df1.index.isna().any():
        df1 = df1[df1.index.notna()]
        warnings.warn("Input dataframe index contains NA values", UserWarning)

    # for duplicated gene index, keep the first one
    if df1.index.duplicated().any():
        df1 = df1[~df1.index.duplicated(keep='first')]
        warnings.warn("Input dataframe contains duplicated indices", UserWarning)

    score_df = pd.DataFrame()
    score_df.index = df1.columns

    if isinstance(method, str):
        method = [method]
    
    for score in method:
        if score == "ESTIMATE":
            score_est = estimateScore.ESTIMATEscore(df1)
            score_df = pd.concat([score_df, score_est], axis=1)
        elif score == "ISTME":
            score_istme = ISM.istmeScore(df1)
            score_df = pd.concat([score_df, score_istme], axis = 1)
        elif score == "NetBio":
            score_nb = netbio.get_netbio(df1)
            score_df = pd.concat([score_df, score_nb], axis = 1)
        elif score == "SIA":
            score_sia = SIAscore.sia_score(df1)
            score_df = pd.concat([score_df, score_sia], axis = 1)
        else:
            raise ValueError("Invalid scoring method")
    
    return score_df


from setuptools import setup, find_packages

VERSION = '1.1.2' 
DESCRIPTION = 'Python package for calculating TME scores'


# Setting up
setup(
        name="TMEImmune", 
        version=VERSION,
        author="Qilu Zhou",
        author_email="<qiluzhou@umass.edu>",
        url="https://github.com/qiluzhou/TMEImmune",
        description=DESCRIPTION,
        long_description=open("README.md").read(),
        long_description_content_type="text/markdown",
        packages=find_packages(),
        install_requires=["pandas>=1.5.0", "numpy>=1.23.5", "cmapPy", "rnanorm", "gseapy == 0.11.0",
                          "inmoose", "lifelines", "scikit-learn", "matplotlib", "requests"], 
        keywords=['python', 'TME score'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ],
        python_requires=">=3.10"
)


import pandas as pd
import numpy as np

def sia_score(df, marker = "C1QA"): # C1QA, C1QB, C1QC
    """
    Compute the SIA score from the provided dataframe
    df: normalized gene expression matrix, with gene symbol as the first column or index
    index: whether gene symbols are the row indices of df
    marker: choose from C1QA, C1QB, C1QC as a biomarker for macrophages, by default the expression of C1QA
    """
    genes = df.index
    sig = ["CD8A", marker]
    # test if the two genes are in the dataframe
    if not all(i in list(genes) for i in sig):
        raise ValueError("Genes needed to get the SIA score do not exist in the input data")
    if any(df.loc[marker, :]) == 0:
        raise ValueError(f"Expression of {marker} is 0")

    sia_score = df.loc['CD8A', :]/df.loc[marker, :]

    sia = pd.DataFrame({"SIA":sia_score})
    sia.index = df.columns

    return sia
